import logging
import os
from google.cloud import storage


logger = logging.getLogger(__name__)
# Ensure load google credentials
# os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = "yaasr/settings/google-cloud-storage-credential.json"


def upload_to_google_cloud_storage(stream_path,
                                   bucket_name,
                                   destination_blob_name=None,
                                   delete_on_success=False):
    """Uploads a file to the Google Cloud storage bucket.
        Ref: https://cloud.google.com/storage/docs/uploading-objects#storage-upload-object-python
        """

    filename = stream_path.split('/')[-1]
    logging.info(f'Uploading {filename} to google cloud storage')
    if destination_blob_name is None:
        destination_blob_name = filename

    storage_client = storage.Client()
    bucket = storage_client.bucket(bucket_name)
    # if bucket don't exists:
    # bucket = storage_client.create_bucket(bucket_name)

    blob = bucket.blob(destination_blob_name)

    blob.upload_from_filename(stream_path)

    logging.info(f'{filename} uploaded to google cloud storage at bucket {bucket_name} as {destination_blob_name}')

    if delete_on_success:
        logging.info(f'Deleting {stream_path}')
        os.remove(stream_path)
    else:
        return stream_path



"""
Iterate over priority domains and send data to server
"""
import argparse
import json
import logging.config
import os
from yaasr import get_all_streams
from yaasr.recorder.stream import YStream
from yaasr.processors.audio.reduce import reformat
from yaasr.processors.archive.google_cloud import upload_to_google_cloud_storage
from yaasr.terminal.supervisor import setup_supervisor


def ls(test_streams=False):
    """ List all available streams """
    # list all streams
    streams = get_all_streams()
    for stream in streams:
        ys = YStream(stream)
        ys.load()
        print(f'{stream}: {ys.streams[0]["url"]}')
        if test_streams:
            ys.record(total_seconds=30, chunk_bytes_size=512, chunk_time_size=1000)


def info(stream):
    """ Shows stream info """
    # list all streams
    ys = YStream(stream)
    ys.load()
    data = json.dumps(ys.data, indent=4)
    print(data)


def record(stream, total_seconds=300, chunk_bytes_size=256, chunk_time_size=60):
    """ Shows stream info """
    # list all streams
    ys = YStream(stream)
    ys.load()
    ys.record(total_seconds=total_seconds, chunk_bytes_size=chunk_bytes_size, chunk_time_size=chunk_time_size)


def compress_and_google_store(stream,
                              total_seconds=300,
                              chunk_bytes_size=256,
                              chunk_time_size=60,
                              audio_format='ogg'):
    """ Shows stream info """

    ys = YStream(stream)
    ys.load()
    ys.post_process_functions = [
        {
            'fn': reformat,
            'params': {
                'audio_format': audio_format,
                'mono': True,
                'delete_on_success': True
            }
        },
        {
            'fn': upload_to_google_cloud_storage,
            'params': {
                'bucket_name': 'parlarispa-radio',
                'delete_on_success': True
            }
        }
    ]
    ys.record(total_seconds=total_seconds, chunk_bytes_size=chunk_bytes_size, chunk_time_size=chunk_time_size)


def main():
    # TODO fix path or replace to local settings
    logging.config.fileConfig('yaasr/log.conf')
    parser = argparse.ArgumentParser()
    parser.add_argument('command', help='Command to run')
    parser.add_argument('--log_level', nargs='?', default='INFO', type=str)
    parser.add_argument('--stream', nargs='?', default=None, type=str)
    # record parameters
    parser.add_argument('--total_seconds', nargs='?', default=0, type=int)
    parser.add_argument('--chunk_bytes_size', nargs='?', default=256, type=int)
    parser.add_argument('--chunk_time_size', nargs='?', default=1200, type=int)
    # compress parameters
    parser.add_argument('--audio_format', nargs='?', default='mp3', choices=['mp3', 'ogg'], type=str)

    # credentials
    parser.add_argument('--google-credentials', nargs='?', default=None, type=str)
    parser.add_argument('--system-user', nargs='?', default=None, type=str)

    args = parser.parse_args()

    if args.google_credentials is not None:
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = args.google_credentials

    if args.command == 'ls':
        return ls()
    elif args.command == 'test':
        return ls(test_stream=True)
    elif args.command == 'info':
        return info(stream=args.stream)
    elif args.command == 'record':
        return record(
            stream=args.stream,
            total_seconds=args.total_seconds,
            chunk_bytes_size=args.chunk_bytes_size,
            chunk_time_size=args.chunk_time_size
        )
    elif args.command == 'compress-and-google-store':
        return compress_and_google_store(
            stream=args.stream,
            total_seconds=args.total_seconds,
            chunk_bytes_size=args.chunk_bytes_size,
            chunk_time_size=args.chunk_time_size,
            audio_format=args.audio_format
        )
    elif args.command == 'supervisor':
        return setup_supervisor(
            stream_name=args.stream,
            system_user_name=args.system_user,
            google_credentials_path=args.google_credentials,
            total_seconds=args.total_seconds,
            chunk_bytes_size=args.chunk_bytes_size,
            chunk_time_size=args.chunk_time_size
        )



import os
from jinja2 import Template
from yaasr import get_all_streams


def setup_supervisor(stream_name,
                     system_user_name,
                     google_credentials_path,
                     total_seconds=0,
                     chunk_bytes_size=256,
                     chunk_time_size=30 * 60):

    """ setup supervisor to save a radio station. """
    streams = get_all_streams() if stream_name.lower() == 'all' else [stream_name]
    template_file = os.path.join(os.path.dirname(__file__), 'tpl', 'supervisor.ini')
    context = {
        'user_name': system_user_name,
        'google_credentials_path': google_credentials_path,
        'total_seconds': total_seconds,
        'chunk_bytes_size': chunk_bytes_size,
        'chunk_time_size': chunk_time_size
    }
    template = Template(open(template_file).read())
    for stream in streams:
        context['stream_name'] = stream
        result = template.render(**context)
        path = f'/etc/supervisor/conf.d/{stream}.conf'
        f = open(path, 'w')
        f.write(result)
        f.close()

        print(f'{path} was created. Run "sudo supervisorctl start yaasr-{stream}" to start recording audios')



import os


__VERSION__ = '0.4.1'
BASE_FOLDER = os.path.dirname(__file__)
STREAMS_FOLDER = os.path.join(BASE_FOLDER, "streams")


def get_all_streams():
    streams = []
    for subdir, dirs, files in os.walk(STREAMS_FOLDER):
        for sdir in dirs:
            data_file = os.path.join(STREAMS_FOLDER, sdir, 'data.json')
            if os.path.isfile(data_file):
                streams.append(sdir)

    return streams



import json
import logging
import os
import requests
from datetime import datetime, timedelta
from yaasr import STREAMS_FOLDER
from yaasr.exceptions import StreamFolderNotFoud, StreamDataFileNotFoud


logger = logging.getLogger(__name__)


class YStream:
    """ YAASR stream """

    def __init__(self, stream_name, streams_folder=STREAMS_FOLDER, destination_folder=''):
        self.name = stream_name
        self.streams_folder = streams_folder
        self.str_chunk_time_format = '%Y-%m-%d--%H-%M-%S'
        # after save each audio chunk we can post-process the file, upload or anything
        self.post_process_functions = []
        self.short_name = self.name
        self.destination_folder = destination_folder

    def get_stream_folder(self):
        """ Get the stream folder """
        stream_folder = os.path.join(self.streams_folder, self.name)
        if not os.path.isdir(stream_folder):
            raise StreamFolderNotFoud(f'Stream not found {stream_folder}')

        return stream_folder

    def load(self):
        """ Load and validate the stream data """
        self.stream_folder = self.get_stream_folder()
        data_file = os.path.join(self.stream_folder, 'data.json')
        if not os.path.isfile(data_file):
            raise StreamDataFileNotFoud(f'Data file not found {data_file}')

        self.data = json.load(open(data_file))

        self.title = self.data['title']
        self.web_site = self.data.get('web', None)
        self.streams = self.data['streams']
        self.short_name = self.data.get('short_name', self.name)

    def generate_stream_path(self, extension):
        now = datetime.now()
        stime = now.strftime(self.str_chunk_time_format)
        stream_path = os.path.join(self.destination_folder, f'{self.short_name}-{stime}.{extension}')
        return now, stream_path

    def record(self, total_seconds=0, chunk_bytes_size=1024, chunk_time_size=60):
        """ Record the online stream

        Params:
            total_seconds: total time to save from the stream. 0 is for ever
            chunk_bytes_size: chunk size to iterate over stream downloaded data
            chunk_time_size: split the audio files is chunk with this time
        """
        c = 0
        for stream in self.streams:

            c += 1
            url = stream['url']
            logger.info(f'Attempt to record from {url}')
            try:
                r = requests.get(url, stream=True)
            except Exception as e:
                logger.error(f'Error connecting to stream {c} {url}: {e}')
                continue

            extension = stream.get('extension', 'mp3')
            start, stream_path = self.generate_stream_path(extension=extension)

            f = open(stream_path, 'wb')
            logger.info(f'Recording from {url}')
            last_start = start
            c = 0
            for block in r.iter_content(chunk_bytes_size):
                c += 1
                f.write(block)
                logger.debug('  ... chunk saved')

                now = datetime.now()
                elapsed = now - start
                if total_seconds > 0 and elapsed >= timedelta(seconds=total_seconds):
                    logger.info(f'Finish recording {now}')
                    break
                elif now - last_start >= timedelta(seconds=chunk_time_size):
                    logger.info(f'{now} Elapsed {elapsed} Finish chunk {c}')
                    self.chunk_finished(stream_path)
                    last_start, stream_path = self.generate_stream_path(extension=extension)
                    f.close()
                    f = open(stream_path, 'wb')

            f.close()
            # last chunk
            self.chunk_finished(stream_path)
            return stream_path

    def chunk_finished(self, stream_path):
        """ An audio chunk finished. We can post-process and/or upload """
        logger.info('Chunk finished')
        for ppf in self.post_process_functions:
            fn = ppf['fn']
            logger.info(f'Running {fn}')
            params = ppf.get('params', {})
            stream_path = fn(stream_path, **params)



"""
Upload through SSH
"""
import logging
import os
from paramiko import SSHClient, AutoAddPolicy


logger = logging.getLogger(__name__)


def upload_ssh(stream_path,
               host,
               user,
               password,
               destination_folder,
               port=22,
               delete_on_success=False):

    """ Upload with SSH """
    # TODO allow use keys instead user/pass
    logging.info(f'Uploading {stream_path}')

    ssh = SSHClient()
    ssh.set_missing_host_key_policy(AutoAddPolicy())
    ssh.connect(host, port, user, password)

    sftp = ssh.open_sftp()

    filename = stream_path.split('/')[-1]
    destination_path = os.path.join(destination_folder, filename)
    logging.info(f'Copying from {stream_path} to {destination_path}')
    sftp.put(localpath=stream_path, remotepath=destination_path)

    sftp.close()
    ssh.close()

    if delete_on_success:
        logging.info(f'Deleting {stream_path}')
        os.remove(stream_path)
    else:
        return stream_path



import logging
import os
from pydub import AudioSegment
from pydub.exceptions import CouldntDecodeError


logger = logging.getLogger(__name__)


def tryopen(stream_path):
    # https://github.com/jiaaro/pydub
    extension = stream_path.split('.')[-1]
    extensions = [extension, 'aac', 'mp4', 'mp3', 'flv', 'ogg', 'wma', 'wav']
    for ext in extensions:
        try:
            audio = AudioSegment.from_file(stream_path, ext)
        except CouldntDecodeError:
            # maybe the codec is not the format
            logger.error
        else:
            logging.info(f'{ext} worked')
            return audio
    raise Exception(f'Pydub can\'t decode {stream_path} from {extensions}')


def reformat(stream_path,
             audio_format='mp3',
             mono=True,
             delete_on_success=False):
    """ change the format and return new format with desired bitrate """
    logging.info(f'reformat {stream_path} to {audio_format}')
    audio = tryopen(stream_path)
    # if mono:
    #     audio = audio.set_channels(1)
    filename = '.'.join(stream_path.split('.')[:-1])
    new_file_name = f'{filename}.{audio_format}'

    if audio_format == 'mp3':
        parameters = [
            "-ab", "16k",
            "-ar", "8000"
        ]
    elif audio_format == 'ogg':
        parameters = [
            "-ab", "16k",
            "-ar", "8000"
        ]

    if mono:
        parameters += ['-ac', '1']

    ret = audio.export(new_file_name, format=audio_format, parameters=parameters)
    logging.info(f'Export finished {ret}')

    if delete_on_success:
        logging.info(f'Deleting {stream_path}')
        os.remove(stream_path)
    return new_file_name



import setuptools
from yaasr import __VERSION__


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setuptools.setup(
    name="yaasr",
    version=__VERSION__,
    author="Andres Vazquez",
    author_email="andres@data99.com.ar",
    description="Yet another audio stream recorder",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/avdata99/yaasr",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers"
    ],
    packages=setuptools.find_packages(),
    package_data={
        "yaasr": [
            "streams/*/data.json",
            "terminal/tpl/supervisor.ini"
            ]
        },
    python_requires='>=3.6',
    install_requires=[
        'requests>=2.25.1',
        'pydub>=0.24.1',   # Process audios
        'paramiko>=2.7.2',  # Upload ssh
        'google-cloud-storage>=1.36.0',  # Google cloud storage
        'jinja2>=2.11.3',  # Supervisor templates creation
    ],
    entry_points={
        'console_scripts': [
            'yaasr=yaasr.terminal:main',
            ],
    },
)
